package NIST;

# ------------------------------------------------------------------------

#Copyright (C) 2005 Enrique Amig and Jess Gimnez

#This library is free software; you can redistribute it and/or
#modify it under the terms of the GNU Lesser General Public
#License as published by the Free Software Foundation; either
#version 2.1 of the License, or (at your option) any later version.

#This library is distributed in the hope that it will be useful,
#but WITHOUT ANY WARRANTY; without even the implied warranty of
#MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#Lesser General Public License for more details.

#You should have received a copy of the GNU Lesser General Public
#License along with this library; if not, write to the Free Software
#Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

# ------------------------------------------------------------------------

use strict;
use Data::Dumper;
use IQ::IQXML;
use IQ::COMMON;
use File::ReadBackwards;

use vars qw($NISTEXT $rNIST $TNIST);

$NIST::NISTEXT = "NIST";
$NIST::TNIST = "mteval-kit/mteval-v11b.pl -n -d 2"; # version v10
$NIST::rNIST = { "$NIST::NISTEXT-1" => 1, "$NIST::NISTEXT-2" => 1, "$NIST::NISTEXT-3" => 1, "$NIST::NISTEXT-4" => 1, "$NIST::NISTEXT-5" => 1 };

sub read_nist {
   #description _ read NIST value from report file
   #param1  _ report filename
   #@return _ NIST score list (n = 1..9)

   my $report = shift;

   my @lnist;
   my $REPORT = File::ReadBackwards->new($report) or die "Couldn't open input file: $report\n";
   my $line = $REPORT->readline;
   while (!($line =~ /^ +NIST:.*/)) { $line = $REPORT->readline; }
   $REPORT->close();
   chomp($line);
   $line =~ s/^ +NIST: +//; $line =~ s/ +/ /g;
   my @lnist = split(" ", $line);
   pop(@lnist);

   return \@lnist;
}

sub read_nist_segments {
   #description _ read NIST-5 value from report file (for all segments)
   #param1  _ report filename
   #@return _ NIST score list

   my $report = shift;

   my @lnist1;
   my @lnist2;
   my @lnist3;
   my @lnist4;
   my @lnist5;

   open(AUX, "< $report") or die "couldn't open file: $report\n";
   while (my $aux = <AUX>) {
      if ($aux =~ /^ +NIST score using.*/) {
         chomp($aux);
         my @laux = split(/ +/, $aux);
         push(@lnist1, $laux[6]);
         push(@lnist2, $laux[7]);
         push(@lnist3, $laux[8]);
         push(@lnist4, $laux[9]);
         push(@lnist5, $laux[10]);
      }
   }
   close(AUX);
   my @SEG = (\@lnist1, \@lnist2, \@lnist3, \@lnist4, \@lnist5);

   return \@SEG;
}

sub computeMultiNIST {
   #description _ computes NIST score (by calling NIST mteval script) -> n = 1..4 (multiple references)
   #param1 _ source file
   #param2 _ candidate file
   #param3 _ reference file(s) [hash reference]
   #param4 _ remake reports? (1 - yes :: 0 - no)
   #param5 _ TARGET NAME (base filename)
   #param6 _ REF NAME (base filename)
   #param7 _ tools
   #param8 _ verbosity (0/1)

   my $src = shift;
   my $out = shift;
   my $Href = shift;
   my $remakeREPORTS = shift;
   my $TGTname = shift;
   my $REFname = shift;
   my $tools = shift;
   my $verbose = shift;

   my $toolNIST = "perl $tools/$NIST::TNIST";

   my $refNISTsgml = "./$COMMON::TMP/".rand($COMMON::NRAND).".$COMMON::REFEXT.$NISTEXT.$COMMON::SGMLEXT";
   if ((!(-e $refNISTsgml)) or $remakeREPORTS) { IQXML::NIST_f_create_mteval_multidoc($Href, $refNISTsgml, 2, $verbose); }
   my $srcNISTsgml = "./$COMMON::TMP/".rand($COMMON::NRAND).".$COMMON::SRCEXT.$NISTEXT.$COMMON::SGMLEXT";
   if ((!(-e $srcNISTsgml)) or $remakeREPORTS) { IQXML::NIST_f_create_mteval_doc($src, $srcNISTsgml, 0, $verbose); }
   my $outNISTsgml = "./$COMMON::TMP/".rand($COMMON::NRAND).".$COMMON::SYSEXT.$NISTEXT.$COMMON::SGMLEXT";
   if ((!(-e $outNISTsgml)) or $remakeREPORTS) { IQXML::NIST_f_create_mteval_doc($out, $outNISTsgml, 1, $verbose); }
   my $reportNIST = "./$COMMON::TMP/".rand($COMMON::NRAND).".$NISTEXT.$COMMON::REPORTEXT";
   if ($verbose > 1) { print STDERR "building $reportNIST...\n"; }
   system("$toolNIST -s $srcNISTsgml -t $outNISTsgml -r $refNISTsgml > $reportNIST");
   if (-e $srcNISTsgml) { system "rm -f $srcNISTsgml"; }
   if (-e $refNISTsgml) { system "rm -f $refNISTsgml"; }
   if (-e $outNISTsgml) { system "rm -f $outNISTsgml"; }
   my $SYS = read_nist($reportNIST);
   my $SEG = read_nist_segments($reportNIST);
   system("rm -f $reportNIST");

   return($SYS, $SEG);
}

sub doMultiNIST {
   #description _ computes NIST score (by calling NIST mteval script) -> n = 1..4 (multiple references)
   #param1  _ source file
   #param2  _ candidate file
   #param3  _ reference file(s) [hash reference]
   #param4  _ remake reports? (1 - yes :: 0 - no)
   #param5  _ TARGET NAME (base filename)
   #param6  _ REF NAME (base filename)
   #param7  _ tools
   #param8  _ metrics to compute
   #param9  _ optional prefix
   #param10 _ verbosity (0/1)

   my $src = shift;
   my $out = shift;
   my $Href = shift;
   my $remakeREPORTS = shift;
   my $TGTname = shift;
   my $REFname = shift;
   my $tools = shift;
   my $M = shift;
   my $prefix = shift;
   my $verbose = shift;

   my $GO = 0; my $i = 0;
   my @mNIST = keys %{$NIST::rNIST};
   while (($i < scalar(@mNIST)) and (!$GO)) { if (exists($M->{$prefix.$mNIST[$i]})) { $GO = 1; } $i++; }

   if ($GO) {
      if ($verbose == 1) { print STDERR "NIST.."; }
      my $reportNIST1xml = "./$TGTname/$REFname/$prefix$NISTEXT-1.$COMMON::XMLEXT";
      my $reportNIST2xml = "./$TGTname/$REFname/$prefix$NISTEXT-2.$COMMON::XMLEXT";
      my $reportNIST3xml = "./$TGTname/$REFname/$prefix$NISTEXT-3.$COMMON::XMLEXT";
      my $reportNIST4xml = "./$TGTname/$REFname/$prefix$NISTEXT-4.$COMMON::XMLEXT";
      my $reportNIST5xml = "./$TGTname/$REFname/$prefix$NISTEXT-5.$COMMON::XMLEXT";

      if ((!(-e $reportNIST1xml) and !(-e $reportNIST1xml.".$COMMON::GZEXT")) or (!(-e $reportNIST2xml) and !(-e $reportNIST2xml.".$COMMON::GZEXT")) or (!(-e $reportNIST3xml) and !(-e $reportNIST3xml.".$COMMON::GZEXT")) or (!(-e $reportNIST4xml) and !(-e $reportNIST4xml.".$COMMON::GZEXT")) or (!(-e $reportNIST5xml) and !(-e $reportNIST5xml.".$COMMON::GZEXT")) or $remakeREPORTS) {
         my ($SYS, $SEG) = NIST::computeMultiNIST($src, $out, $Href, $remakeREPORTS, $TGTname, $REFname, $tools, $verbose);
         IQXML::write_report($TGTname, $REFname, "$prefix$NISTEXT-1", $SYS->[0], $SEG->[0], $reportNIST1xml, $verbose);
         IQXML::write_report($TGTname, $REFname, "$prefix$NISTEXT-2", $SYS->[1], $SEG->[1], $reportNIST2xml, $verbose);
         IQXML::write_report($TGTname, $REFname, "$prefix$NISTEXT-3", $SYS->[2], $SEG->[2], $reportNIST3xml, $verbose);
         IQXML::write_report($TGTname, $REFname, "$prefix$NISTEXT-4", $SYS->[3], $SEG->[3], $reportNIST4xml, $verbose);
         IQXML::write_report($TGTname, $REFname, "$prefix$NISTEXT-5", $SYS->[4], $SEG->[4], $reportNIST5xml, $verbose);
      }
   }
}

sub computeNIST {
   #description _ computes NIST score (by calling NIST mteval script) -> n = 1..4 (multiple references)
   #param1 _ source file
   #param2 _ candidate file
   #param3 _ reference file(s) [hash reference]
   #param4 _ remake reports? (1 - yes :: 0 - no)
   #param5 _ TARGET NAME (base filename)
   #param6 _ REF NAME (base filename)
   #param7 _ tools
   #param8 _ verbosity (0/1)

   my $src = shift;
   my $out = shift;
   my $ref = shift;
   my $remakeREPORTS = shift;
   my $TGTname = shift;
   my $REFname = shift;
   my $tools = shift;
   my $verbose = shift;

   my $toolNIST = "perl $tools/$NIST::TNIST";

   my $refNISTsgml = "./$COMMON::TMP/".rand($COMMON::NRAND).".$COMMON::REFEXT.$NISTEXT.$COMMON::SGMLEXT";
   if ((!(-e $refNISTsgml)) or $remakeREPORTS) { IQXML::NIST_f_create_mteval_doc($ref, $refNISTsgml, 2, $verbose); }
   my $srcNISTsgml = "./$COMMON::TMP/".rand($COMMON::NRAND).".$COMMON::SRCEXT.$NISTEXT.$COMMON::SGMLEXT";
   if ((!(-e $srcNISTsgml)) or $remakeREPORTS) { IQXML::NIST_f_create_mteval_doc($src, $srcNISTsgml, 0, $verbose); }
   my $outNISTsgml = "./$COMMON::TMP/".rand($COMMON::NRAND).".$COMMON::SYSEXT.$NISTEXT.$COMMON::SGMLEXT";
   if ((!(-e $outNISTsgml)) or $remakeREPORTS) { IQXML::NIST_f_create_mteval_doc($out, $outNISTsgml, 1, $verbose); }
   my $reportNIST = "./$COMMON::TMP/".rand($COMMON::NRAND).".$NISTEXT.$COMMON::REPORTEXT";
   if ($verbose > 1) { print STDERR "building $reportNIST...\n"; }
   system("$toolNIST -s $srcNISTsgml -t $outNISTsgml -r $refNISTsgml > $reportNIST");
   if (-e $srcNISTsgml) { system "rm -f $srcNISTsgml"; }
   if (-e $refNISTsgml) { system "rm -f $refNISTsgml"; }
   if (-e $outNISTsgml) { system "rm -f $outNISTsgml"; }
   my $SYS = read_nist($reportNIST);
   my $SEG = read_nist_segments($reportNIST);
   system("rm -f $reportNIST");

   return($SYS, $SEG);
}

sub doNIST {
   #description _ computes NIST score (by calling NIST mteval script) -> n = 1..4 (single reference)
   #param1  _ source file
   #param2  _ candidate file
   #param3  _ reference file(s) [hash reference]
   #param4  _ remake reports? (1 - yes :: 0 - no)
   #param5  _ TARGET NAME (base filename)
   #param6  _ REF NAME (base filename)
   #param7  _ tools
   #param8  _ metrics to compute
   #param9  _ optional prefix
   #param10 _ verbosity (0/1)

   my $src = shift;
   my $out = shift;
   my $ref = shift;
   my $remakeREPORTS = shift;
   my $TGTname = shift;
   my $REFname = shift;
   my $tools = shift;
   my $M = shift;
   my $prefix = shift;
   my $verbose = shift;

   my $GO = 0; my $i = 0;
   my @mNIST = keys %{$NIST::rNIST};
   while (($i < scalar(@mNIST)) and (!$GO)) { if (exists($M->{$prefix.$mNIST[$i]})) { $GO = 1; } $i++; }

   if ($GO) {
      if ($verbose == 1) { print STDERR "NIST.."; }
      my $reportNIST1xml = "./$TGTname/$REFname/$prefix$NISTEXT-1.$COMMON::XMLEXT";
      my $reportNIST2xml = "./$TGTname/$REFname/$prefix$NISTEXT-2.$COMMON::XMLEXT";
      my $reportNIST3xml = "./$TGTname/$REFname/$prefix$NISTEXT-3.$COMMON::XMLEXT";
      my $reportNIST4xml = "./$TGTname/$REFname/$prefix$NISTEXT-4.$COMMON::XMLEXT";
      my $reportNIST5xml = "./$TGTname/$REFname/$prefix$NISTEXT-5.$COMMON::XMLEXT";

      if ((!(-e $reportNIST1xml) and !(-e $reportNIST1xml.".$COMMON::GZEXT")) or (!(-e $reportNIST2xml) and !(-e $reportNIST2xml.".$COMMON::GZEXT")) or (!(-e $reportNIST3xml) and !(-e $reportNIST3xml.".$COMMON::GZEXT")) or (!(-e $reportNIST4xml) and !(-e $reportNIST4xml.".$COMMON::GZEXT")) or (!(-e $reportNIST5xml) and !(-e $reportNIST5xml.".$COMMON::GZEXT")) or $remakeREPORTS) {
         my ($SYS, $SEG) = NIST::computeNIST($src, $out, $ref, $remakeREPORTS, $TGTname, $REFname, $tools, $verbose);
         IQXML::write_report($TGTname, $REFname, "$prefix$NISTEXT-1", $SYS->[0], $SEG->[0], $reportNIST1xml, $verbose);
         IQXML::write_report($TGTname, $REFname, "$prefix$NISTEXT-2", $SYS->[1], $SEG->[1], $reportNIST2xml, $verbose);
         IQXML::write_report($TGTname, $REFname, "$prefix$NISTEXT-3", $SYS->[2], $SEG->[2], $reportNIST3xml, $verbose);
         IQXML::write_report($TGTname, $REFname, "$prefix$NISTEXT-4", $SYS->[3], $SEG->[3], $reportNIST4xml, $verbose);
         IQXML::write_report($TGTname, $REFname, "$prefix$NISTEXT-5", $SYS->[4], $SEG->[4], $reportNIST5xml, $verbose);
      }
   }
}

1;
