# The METEOR automatic MT evaluation metric. See README.txt for more information.

# License Start: 
#                    Carnegie Mellon University                      
#                      Copyright (c) 2004                            
#                       All Rights Reserved.                         
#
# Permission is hereby granted, free of charge, to use and distribute
# this software and its documentation without restriction, including 
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of this work, and to    
# permit persons to whom this work is furnished to do so, subject to 
# the following conditions:                                          
#  1. The code must retain the above copyright notice, this list of  
#     conditions and the following disclaimer.                       
#  2. Any modifications must be clearly marked as such.              
#  3. Original authors' names are not deleted.                       
#  4. The authors' names are not used to endorse or promote products 
#     derived from this software without specific prior written      
#     permission.                                                    
#
# CARNEGIE MELLON UNIVERSITY AND THE CONTRIBUTORS TO THIS WORK       
# DISCLAIM ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING    
# ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT 
# SHALL CARNEGIE MELLON UNIVERSITY NOR THE CONTRIBUTORS BE LIABLE    
# FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES  
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN 
# AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,        
# ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF     
# THIS SOFTWARE.                                                     
#
# Author: Satanjeev "Bano" Banerjee satanjeev@cmu.edu
# Author: Alon Lavie alavie@cs.cmu.edu
#
# License End.

use mStageMatcher; # multi stage matcher algorithm 
use Getopt::Long; # for processing the command line

# The following file will be used as a stop list if neither the
# -noStop switch is used nor an alternate stop list is provided with
# the -stop switch
my $DEFAULT_STOP_LIST = "default_stop_list.txt"; 

# If no options have been provided provide the usage notes 
printUsage() if ($#ARGV == -1);

# If options have been provided, read them in
GetOptions("l=s", "t=s", "r=s", "s=s", "noStop", "stop=s", "modules=s");

######################################################################################
#####  Main routine to first handle the command line options, then drive the rest of 
#####  the program.
######################################################################################

my $doLog = 0;
if (defined $opt_l) {
  $doLog = 1;

  # open the log file for detailed output, if debug level is non zero
  open LOG, ">$opt_l" || die "Couldn't open log file $opt_l for writing\n";
}

# initialize stop-list hash
my %stopHash = ();
my $stopListFile = "";
if (defined $opt_noStop) {
  if (defined $opt_stop) {
    die "Cannot simultaneously use both the --noStop switch as well as the --stop switch.\n"
  }
} else {
  if (defined $opt_stop) {
    $stopListFile = $opt_stop;
    print "Using $stopListFile as the stop list.\n";
  } else {
    $stopListFile = findStopListFile($DEFAULT_STOP_LIST);
    if ($stopListFile ne "") { 
	  print "Using $stopListFile as the default stop list.\n"; 
	} else { 
	  print "Warning: Cannot find default stop list ($DEFAULT_STOP_LIST) in your library path - proceeding without any stop lists.. To avoid this warning either (a) use --noStop to specify no stop list, (b) use --stop <file> to provide a stop list, or (c) add the path to the default stop list (should be where you unzipped METEOR) to your PERL5LIB environment variable.\n"; 
	}
  }
}

if ($stopListFile ne "") {
  open (STOP, $stopListFile) || die "Couldn't open stop file $opt_stop - aborting.\n";
  while (<STOP>) {
    chomp;
    s/\s//g; # get rid of all the random \n, \r stuff that chomp didn't get rid of
    $stopHash{$_} = 1;
  }
  close(STOP);
}

my @modules = ();
if (defined $opt_modules) {
  @modules = split /\s+/, $opt_modules;
} else {
  print "Modules list not provided... using default modules \"exact porter_stem\"\n";
  @modules = qw(exact porter_stem);
}

# Check the list of modules to see if we need to initialize WordNet
foreach (@modules) {
  if (/^wn_/) { # WordNet based modules are assumed to start with "wn_"
    # include Jason Rennie's perl module that will interface with WordNet
    require WordNet::QueryData; 
    # create a WordNet object, this will later be put in the input hash
    $wn = WordNet::QueryData->new;
    last; 
  }
}

# Initialize data structures that will hold all reference and hypothesis files
# Data structure for references and hypotheses:
# 
# Outer hash: key = "docid::segid", value = inner hash
# Inner hash: key = "ref-name/hyp-name", value = ref/hyp
# This should be able to handle any number of references as well as any number of hypotheses!

my %allData = ();     # all the hypothesis and reference strings 
my %systemIDs = ();   # all the automatic system ids (found in the testing file) 
my %humanRefIDs = (); # all the reference ids (found in the reference file)

# Next read in the test and reference files, abort if not supplied
if (defined $opt_t) {
  readSGMfile($opt_t, \%allData, \%systemIDs);
} else {
  print "Please specify a test file using the -t option.\n";
  printUsage();
}
    
if (defined $opt_r) { 
  readSGMfile($opt_r, \%allData, \%humanRefIDs);
} else {
  print "Please specify a reference file using the -r option.\n";
  printUsage();
}
    
# Check if provided system id belongs to the list of system ids found
my $givenSystemID = "";
if (defined $opt_s) {
  $givenSystemID = $opt_s;
} else {
  print "Please specify a system id using the -s option.\n";
  printUsage();
}
    
die "Input system id not found in given test file\n"
  unless (defined $systemIDs{$givenSystemID});
    
# Check the data for holes. That is, check to ensure that every
# document::seg-id pair has exactly one sentence from each of the
# system ids as well as the human ids
if (!(dataComplete(\%allData, \%systemIDs, \%humanRefIDs))) {
  print "Data looks incomplete - aborting.\n";
}

# # debug printing     
# foreach my $outerKey (sort keys %allData) {
#   print "Key: $outerKey\n";
#   foreach my $innerKey (sort keys %{$allData{$outerKey}}) {
#     print "  Inner key: $innerKey\n";
#   }
# }
# print "The system ids are:\n";
# foreach my $systemID (sort keys %systemIDs) {
#   print "    $systemID\n";
# }
# print "The human reference ids are:\n";
# foreach my $humanRefID (sort keys %humanRefIDs) {
#   print "    $humanRefID\n";
# }

# now score each hypothesis, and from that, create an aggregate score
# for this system
my $numSegmentsScored = 0;
print "\nSegment by segment results:\n\n";
my %aggregateScoresHash = getSystemWideScore(\$numSegmentsScored, %allData);

# print overall output
print "Overall system score:\n";    
print getLogString(%aggregateScoresHash);
print "Number of segments scored: $numSegmentsScored\n";

# done!!


######################################################################################
#####  Sub: getSystemWideScore(%allData)
#####  
#####  Subroutine to compute hypothesis by hypothesis score, and from those, the
#####  overall system score
######################################################################################

sub getSystemWideScore {
  my $numSegmentsScoredRef = shift;
  my %allData = @_;

  # initialize necessary overall (system wide) metrics
  my %aggregateScoresHash = ();
  $aggregateScoresHash{"totalMatches"} = 0;
  $aggregateScoresHash{"numChunks"} = 0;
  $aggregateScoresHash{"hypLength"} = 0;
  $aggregateScoresHash{"refLength"} = 0;
  ${$numSegmentsScoredRef} = 0;

  # now go through every segment in every document. 
  foreach my $outerKey (sort keys %allData) {
    die "Internal error during scoring hypothesis\n" 
      unless ($outerKey =~ /(.*)::(.*)/);
    my $documentID = $1;
    my $segmentID = $2;
    
    print "Document: $documentID, Segment: $segmentID\n";

    my $hypothesis = "";
    if (defined ${$allData{$outerKey}}{$givenSystemID}) {
      $hypothesis = ${$allData{$outerKey}}{$givenSystemID};
    } else {
      die "Could not find hypothesis translation for document $documentID, segment $segmentID\n";
    }
    
    # do some low level clean up of hypothesis
    normalizeText(\$hypothesis);
    
    # now compute the best score for this hypothesis, across all the references
    my %bestScore = (); 
    
    # compare the hypothesis translation with each reference translation. 
    foreach my $referenceID (sort keys %humanRefIDs) {
      # we will compute a whole bunch of metrics after comparing the
      # hyp witth this particular reference. We'll put them all into the
      # "scoresHash"
      my %scoresHash = ();
      
      my $reference = "";
      if (defined ${$allData{$outerKey}}{$referenceID}) {
	$reference = ${$allData{$outerKey}}{$referenceID};
      } else {
	die "Could not find reference translation with reference ID $referenceID in document $documentID, segment $segmentID\n";
      }
      
      # do some low level clean up of hypothesis
      normalizeText(\$reference);

      # use mStageMatcher.pm to do the string matching 
      # mStageMatcher.pm takes a hash as input. Construct the hash.
      my %inputHash = ();
      
      # put in the two strings
      $inputHash{"firstString"} = $hypothesis;
      $inputHash{"secondString"} = $reference;
      
      # set maxComputations
      $inputHash{"maxComputations"} = 10000;
      
      # put in the modules array
      @{$inputHash{"modules"}} = @modules; # this array comes from the commandline
      
      # pass in stop hash (maybe an empty one!)
      %{$inputHash{"stop"}} = %stopHash;
      
      # Make sure pruning is set to "on"
      $inputHash{"prune"} = 1;
      
      # put in the detail flag to find out number of computations being done
      $inputHash{"details"} = 1;
      
      # if $wn is defined, send the WordNet object along
      $inputHash{"wn"} = $wn if (defined $wn);
      
      # Now call the actual matching function (this is defined in mStageMatcher.pm
      match(\%inputHash);
      
      # print number of modules
      print LOG $#{$inputHash{"modules"}} + 1, "\n" if ($doLog);
      
      # Count up the total number of matches across all the stages
      $scoresHash{"totalMatches"} = 0;
      for (my $i = 0; $i <= $#{$inputHash{"matchScore"}}; $i++) {
		my $numMatches = ${$inputHash{"matchScore"}}[$i][0];
		my $numFlips = ${$inputHash{"matchScore"}}[$i][1];
		
		print LOG "$numMatches $numFlips\n" if ($doLog);
		$scoresHash{"totalMatches"} += $numMatches;
      }
      
      # print the num chunks and avg chunk lengh
      $scoresHash{"numChunks"} = $inputHash{"numChunks"};
      print LOG $inputHash{"numChunks"}, " ", $inputHash{"avgChunkLength"}, "\n" if ($doLog);

      print "  Reference: $referenceID, ";
      print "Matches found: " . $scoresHash{"totalMatches"} . ", ";
      print "Chunks found: " . $scoresHash{"numChunks"} . "\n";
      
      # get the lengths of the two strings
      $scoresHash{"hypLength"} = $inputHash{"sizeOfFirstString"};
      $scoresHash{"refLength"} = $inputHash{"sizeOfSecondString"};
      
      # compute the score and all the other metrics for this scores hash
      # for this hypothesis - reference pair
      # When total matches == 0, then score is defined to be 0, and all
      # other measures are undefined.
      computeMetrics(\%scoresHash);
      print "sgm", getLogString(%scoresHash);  #this line has been MODIFIED by IQMT developers
      
      # update info for best score, if need be
      %bestScore = %scoresHash 
        if ((!(defined $scoresHash{"score"})) || ($scoresHash{"score"} >= $bestScore{"score"}));
    }
    
    # print a new line after all the references are done for this hypothesis
    print "\n";
    
    # print best score to log file
    if ($doLog) {
      print LOG "doc_Id,$documentID\n";
      print LOG "Segment ID,$segmentID\n";
      print LOG getLogString(%bestScore);
    }
	
    # update aggregate statistics
    $aggregateScoresHash{"totalMatches"} += $bestScore{"totalMatches"};
    $aggregateScoresHash{"numChunks"} += $bestScore{"numChunks"};
    $aggregateScoresHash{"hypLength"} += $bestScore{"hypLength"};
    $aggregateScoresHash{"refLength"} += $bestScore{"refLength"};
	${$numSegmentsScoredRef}++;
  }
  
  # compute metrics over the aggregate scores
  computeMetrics(\%aggregateScoresHash);
  return (%aggregateScoresHash);
}



# subroutine to take a scoresHash as an input, and to return a string
# suitable for printing out to the log output
sub getLogString {
  my %scoresHash = @_;
  
  my @labels = ({"Score" => "score"},
				{"Matches" => "totalMatches"},
				{"Chunks" => "numChunks"},
				{"HypLength" => "hypLength"},
				{"RefLength" => "refLength"},
				{"Precision" => "precision"}, 
				{"Recall" => "recall"},
				{"1-Factor" => "f1"},
				{"Fmean" => "fmean"},
				{"Penalty" => "penalty"},
				{"Fragmentation" => "frag"}
			   );
  
  my $outputString = "";
  foreach my $hash (@labels) {
    foreach my $label (keys %{$hash}) {
      my $value = ${$hash}{$label};
      $outputString .= $label;
      $outputString .= ": ";
      $outputString .= ((defined $scoresHash{$value}) ? (sprintf "%.4f", $scoresHash{$value}) : "undefined");	
      $outputString .= "\n";
    }
  }
  
  return $outputString;
}



# subroutine to compute all score metrics, given totalMatches, chunks,
# hypLength and refLength
sub computeMetrics {
  my $scoresHashRef = shift;
  
  if (${$scoresHashRef}{"totalMatches"} == 0) {
    ${$scoresHashRef}{"score"} = 0;
  } else {
    # compute precision, recall, f1 and fmean
    ${$scoresHashRef}{"precision"} = ${$scoresHashRef}{"totalMatches"} / ${$scoresHashRef}{"hypLength"};
    ${$scoresHashRef}{"recall"} = ${$scoresHashRef}{"totalMatches"} / ${$scoresHashRef}{"refLength"};
    ${$scoresHashRef}{"f1"} = 2 * ${$scoresHashRef}{"precision"} * ${$scoresHashRef}{"recall"} / 
      (${$scoresHashRef}{"precision"} + ${$scoresHashRef}{"recall"});
    ${$scoresHashRef}{"fmean"} = 1 / ((1 / 10) * ((1 / ${$scoresHashRef}{"precision"}) + 
						  (9 / ${$scoresHashRef}{"recall"}))
				     );
    
    # compute fragmentation and penalty
    if (${$scoresHashRef}{"totalMatches"} == 1) {
      ${$scoresHashRef}{"frag"} = 1;
      ${$scoresHashRef}{"penalty"} = 0;
    } else {
      ${$scoresHashRef}{"frag"} = ${$scoresHashRef}{"numChunks"} / ${$scoresHashRef}{"totalMatches"};
      ${$scoresHashRef}{"penalty"} = 0.5 * (${$scoresHashRef}{"frag"} ** 3);
    }
    
    # compute score based on fmean fragmentation and pentaly. 
    ${$scoresHashRef}{"score"} = ${$scoresHashRef}{"fmean"} * (1 - ${$scoresHashRef}{"penalty"});
  }
}



######################################################################################
#####  Sub: readSGMfile($fileName, $allDataRef, $idRef)
#####  
#####  Subroutine to read in all the hypothesis/references in the file with the given
#####  "filename". Hypothesis data will be filled into the hash with the given reference
#####  "allDataRef" and all the system ids or human ids that are found in this file will
#####  be pushed into the hash with the given reference $idRef
##### 
######################################################################################

sub readSGMfile {
  my $fileName = shift;
  my $allDataRef = shift;
  my $idRef = shift;

  open (FILE, $fileName) || die "Couldn't open $fileName\n";
  my $currentDocID = "";
  my $currentSysID = "";
  my $lastAutoSegID = "";
  
  while (my $line = <FILE>) {
    # Example: <DOC docid="AFA20030101.5900" sysid="ahd">
    # Note: In Tides 2002 data, different systems had tags with different cases (docid vs 
    # docID), so in all regexes below, do case insensitive matching (using the "i" flag)
    if ($line =~ /<DOC/i) {
      if ($line =~ /docid=\"([^\"]*)\"/i) {
	$currentDocID = $1;
      } else {
	die "Couldn't read document id from line: $line\n";
      }
      if ($line =~ /sysid=\"([^\"]*)/i) {
	$currentSysID = $1;
      } else {
	die "Couldn't read segment id from line: $line\n";
      }

      $lastAutoSegID = 0;
      # save the system id
      ${$idRef}{$currentSysID} = 1;
      next;
    }
  
    if ($line =~ /<\/DOC>/i) {
      $currentDocID = "";
      $currentSysID = "";
      $lastAutoSegID = 0; # redundant with re-initialization within the if (/<DOC... block, but what the heck!
      next; 
    }

    # Example: <seg id=1> Izzet Ibrahim Meets Saudi Trade Official in Baghdad </seg>
    # note that it is possible that the opening and closing seg tags are on separate lines
    if ($line =~ /<[^>]*?seg[^>]*?id[^>]*?=[^>]*?(\d+)[^>]*?>/i) {
      my $currentSegID = $1;
      while (($line !~ /<\/seg>/i) && (my $nextLine = <FILE>)) {
        $line .= $nextLine;
      }
      my $string = "";
      $lineToMatch = $line;
      $lineToMatch =~ s/\s/ /g;
      if ($lineToMatch =~ /<[^>]*?seg[^>]*?id[^>]*?=[^>]*?\d+[^>]*?>(.*)<[^>]*?\/seg[^>]*?>/i) {
		$string = $1;
      } else {
		die "Having trouble reading in the segment information from:\n$line\n";
      }
      
      # if not inside a document, something's gone wrong
      if ($currentDocID eq "") {
		die "The following segment line in file $fileName seems to be outside a DOC block!: [$line]\n";
      }
	  
      # otherwise, create the key, and push the data in!
      my $outerHashKey = "$currentDocID" . "::" . "$currentSegID";
      ${${$allDataRef}{$outerHashKey}}{$currentSysID} = $string;
	  
      next;
    }
    
    # Example: <seg> Izzet Ibrahimi Meets </sseg> 
    # this is the situation when the files don't have seg-ids. In
    # this case, assume seg-ids start with 1 for a new document,
    # and then increase by 1 until the end of that document
    if ($line =~ /<seg>(.*)<\/seg>/i) {
      my $currentSegID = $lastAutoSegID + 1;
      $lastAutoSegID++;
      my $string = $1;
      
      my $outerHashKey = "$currentDocID" . "::" . "$currentSegID";
      ${${$allDataRef}{$outerHashKey}}{$currentSysID} = $string;
      next;
    }
  }
  close(FILE);
}

# subroutine to go through the allData data structure to ensure that
# every document::segment-id pair has exactly one sentence for each of
# the system and human ref ids
sub dataComplete {
  my $allDataRef = shift;
  my $systemIDsRef = shift;
  my $humanRefIDsRef = shift;

  foreach my $outerKey (keys %{$allDataRef}) {
    die "Internal error during checking for data completeness\n" 
      unless ($outerKey =~ /(.*)::(.*)/);
    my $documentID = $1;
    my $segmentID = $2;

    foreach my $systemID (keys %{$systemIDsRef}) {
      if (!(defined ${${$allDataRef}{$outerKey}}{$systemID})) {
	die "No translation found from system $systemID for segment $segmentID in document $documentID\n";
      }
    }
  }

  return 1;
}
    
sub normalizeText {
  my $strPtr = shift;
  
  # language-independent part:
  $$strPtr =~ s/^\s+//;
  $$strPtr =~ s/\n/ /g; # join lines
  $$strPtr =~ s/(\d)\s+(\d)/$1$2/g;  #join digits
  
  # language-dependent part (assuming Western languages):
  $$strPtr =~ tr/[A-Z]/[a-z]/;
  $$strPtr =~ s/\-/ /g;
  $$strPtr =~ s/[\(\)\"\'\.\$,\%\;\/]/ /g; #KS 200204
  
  $$strPtr =~ s/[^a-z0-9 ]/ /g; #KS 200204
  
  $$strPtr =~ s/\s+/ /g; # one space only between words
  $$strPtr =~ s/^\s+//;  # no leading space
  $$strPtr =~ s/\s+$//;  # no trailing space
}

# Subroutine to look for the default stop list file in the directories in the @INC.
sub findStopListFile {
  my $stopListFileName = shift;
  foreach my $dir (@INC) {
	if (-e "$dir/$stopListFileName") {
	  return "$dir/$stopListFileName";
	}
  }
  return "";
}

# Subroutine to print the usage notes for this program.
sub printUsage {
  print "USAGE: perl meteor.pl -s <system2test> -t <test_file> -r <reference_file> [--modules module1 [module2...]] [--noStop | --stop FILE] [-l <output log file>]\n";
  exit;
}
